import numpy as np

from depynd.information import mutual_information


def mrmr(X, y, lamb=0.0, **kwargs):
    """Select effective features in X on predinting y using minimum redundancy maximum relevance feature selection [1]_.

    Parameters
    ----------
    X : array-like, shape (n_samples, d)
        Features.
    y : array-like, shape (n_samples)
        Target. Can be either continuous or discrete.
    lamb: float
        Threshold for independence tests.
    kwargs : dict, default None
        Optional parameters for MI estimation.

    Returns
    -------
    indices : list
        Indices for the selected features.

    References
    ----------
    .. [1] Peng, Hanchuan, Fuhui Long, and Chris Ding. "Feature selection based on mutual information criteria of
    max-dependency, max-relevance, and min-redundancy." IEEE Transactions on pattern analysis and machine intelligence
    27.8 (2005): 1226-1238.
    """
    n, d = X.shape
    selected = []
    while True:
        max_obj = -np.inf
        not_selected = set(range(d)) - set(selected)
        for i in not_selected:
            x = X[:, i]
            rel = mutual_information(x, y, **kwargs)
            red = [mutual_information(x, X[:, j], **kwargs) for j in selected]
            obj = rel - (np.mean(red) if red else 0)
            if max_obj < obj:
                max_obj = obj
                max_idx = i

        if max_obj <= lamb:
            return selected

        selected.append(max_idx)



import numpy as np
from sklearn.preprocessing import scale


def _instability(X, estimator, alpha, ratio, rep_num):
    n, p = X.shape
    b = int(ratio * n)
    indices = [np.random.choice(np.arange(n), size=b) for _ in range(rep_num)]
    samples = [scale(sample) for sample in X[indices, :]]
    adjs = [estimator(sample, alpha) for sample in samples]
    theta = np.sum(adjs, axis=0) / rep_num
    xi = 2 * theta * (1 - theta)
    d = np.sum(xi) / p / (p - 1)
    return d


def stars(X, estimator, lambdas, beta, ratio, rep_num, verbose=False):
    for i, lamb in enumerate(lambdas):
        instability = _instability(X, estimator, lamb, ratio=ratio, rep_num=rep_num)
        if instability > beta:
            return lambdas[i - 1]
        if verbose:
            print('[stars] lambda: %f, instability: %f' % (lamb, instability))
    return 0



import numpy as np
from sklearn.preprocessing import scale
from sklearn.covariance import graph_lasso

from depynd.markov_networks import skeptic, stars, gsmn, iamb, gsmple


def _graphical_lasso(X, lamb):
    cov = np.cov(scale(X), rowvar=False)
    pre = graph_lasso(cov, alpha=lamb)[1]
    adj = ~np.isclose(pre, 0)
    adj[np.eye(len(adj), dtype=bool)] = 0
    return adj


def select(X, method='glasso', criteria='stars', lambdas=None, verbose=False, return_lambda=False, **kwargs):
    if method == 'glasso':
        estimator = _graphical_lasso
    elif method == 'skeptic':
        estimator = skeptic
    elif method == 'gsmn':
        estimator = gsmn
    elif method == 'iamb':
        estimator = iamb
    elif method == 'gsmple':
        estimator = gsmple
    else:
        raise NotImplementedError('Method %s is not implemented.' % method)

    if lambdas is None:
        lambdas = [1e-5, 1e-4, 1e-3, 5e-3, 0.01, 0.03, 0.05, 0.08, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1]
    lambdas = sorted(lambdas, reverse=True)  # sort by descending order

    n, d = X.shape
    if criteria == 'stars':
        beta = kwargs.get('beta', 0.1)
        ratio = kwargs.get('ratio', 10 * (n ** -0.5) if n > 144 else 0.8)
        rep_num = kwargs.get('rep_num', 20)
        lamb = stars(X, estimator, beta=beta, ratio=ratio, rep_num=rep_num, lambdas=lambdas, verbose=verbose)
    else:
        raise NotImplementedError('Criteria %s is not implemented.' % criteria)

    if return_lambda:
        return estimator(X, lamb), lamb
    else:
        return estimator(X, lamb)



import numpy as np

from depynd.information import conditional_mutual_information


def gsmn(X, lamb=0.0, **kwargs):
    """Search structure of an MRF using glow-shrink Markov networks algorithm [1]_.

    Parameters
    ----------
    X : array, shape (n_samples, d)
        Observations of variables.
    lamb: float
        Threshold for independence tests.
    options : dict, default None
        Optional parameters for MI estimation.
    Returns
    ----------
    adj : array, shape (d, d)
        Estimated structure of an MRF.
    References
    ----------
    .. [1] Bromberg, Facundo, Dimitris Margaritis, and Vasant Honavar. "Efficient Markov network structure discovery
        using independence tests." Journal of Artificial Intelligence Research 35 (2009): 449-484.
    """
    n, d = X.shape
    adj = np.zeros([d, d], dtype=bool)
    for i in range(d):
        adj = _grow(adj, i, X, lamb, **kwargs)
        adj = _shrink(adj, i, X, lamb, **kwargs)
    return adj


def _grow(adj, i, X, lamb, **kwargs):
    n, d = X.shape
    x = X[:, i]
    updated = True
    while updated:
        updated = False
        non_adj = ~adj[i] & (np.arange(d) != i)
        for j in non_adj.nonzero()[0]:
            y = X[:, j]
            z = X[:, adj[i]]
            cmi = conditional_mutual_information(x, y, z, **kwargs)
            if cmi > lamb:
                adj[i, j] = adj[j, i] = 1
                updated = True
                break
    return adj


def _shrink(adj, i, X, lamb, **kwargs):
    n, d = X.shape
    x = X[:, i]
    updated = True
    while updated:
        updated = False
        for j in adj[i].nonzero()[0]:
            other_adj = adj[i] & (np.arange(d) != j)
            y = X[:, j]
            z = X[:, other_adj]
            cmi = conditional_mutual_information(x, y, z, **kwargs)
            if cmi <= lamb:
                adj[i, j] = adj[j, i] = 0
                updated = True
                break
    return adj



from depynd.markov_networks._jose import jose
from depynd.markov_networks._skeptic import skeptic
from depynd.markov_networks._gsmn import gsmn
from depynd.markov_networks._iamb import iamb
from depynd.markov_networks._gsmple import gsmple
from depynd.markov_networks._stars import stars
from depynd.markov_networks._select import select

__all__ = ['jose', 'skeptic', 'gsmn', 'iamb', 'stars', 'select', 'gsmple']



import numpy as np

from depynd.information import conditional_mutual_information


def iamb(X, lamb=0.0, **kwargs):
    """Search Markov blanket in a Bayesian network using Incremental Association Markov Blanket algorithm [1]_.

    Parameters
    ----------
    X : array-like, shape (n_samples, d)
        Observations of variables.
    lamb: float
        Threshold for independence test.
    kwargs : dict, default None
        Optional parameters for MI estimation.

    Returns
    -------
    adj : array, shape (d, d)
        Estimated Markov blanket.

    References
    ----------
    .. [1] Tsamardinos, Ioannis, et al. "Algorithms for Large Scale Markov Blanket Discovery." FLAIRS conference. Vol.
    2. 2003.
    """
    n, d = X.shape
    adj = np.zeros([d, d], dtype=bool)
    for i in range(d):
        x = X[:, i]
        updated = True
        while updated:
            vmax = -np.inf
            updated = False
            z = X[:, adj[i]]
            non_adj = ~adj[i] & (np.arange(d) != i)
            for j in non_adj.nonzero()[0]:
                y = X[:, j]
                cmi = conditional_mutual_information(x, y, z, **kwargs)
                if vmax < cmi:
                    vmax = cmi
                    imax, jmax = i, j
            if vmax >= lamb:
                adj[imax, jmax] = adj[jmax, imax] = 1
                updated = True

        for j in adj[i].nonzero()[0]:
            other_adj = adj[i] & (np.arange(d) != j)
            y = X[:, j]
            z = X[:, other_adj]
            cmi = conditional_mutual_information(x, y, z, **kwargs)
            if cmi <= lamb:
                adj[i, j] = adj[j, i] = 0
            if np.count_nonzero(adj[i]) == 0:
                break

    return adj



import numpy as np
from sklearn.utils.validation import check_array

from depynd.information import mi_dr, mi_knn


def mutual_information(X, Y, force_non_negative=False, **kwargs):
    """Estimate mutual information for discrete-continuous mixutres.

    Parameters
    ----------
    X : array-like, shape (n_samples, d_x) or (n_samples)
        The observations of a variable.
    Y : array-like, shape (n_samples, d_y) or (n_samples)
        The observations of the other variable.
    force_non_negative : bool
        If True, the result will be taken max with zero. Default is False.
    kwargs : dict
        Optional parameters for MI estimation.

    Returns
    -------
    mi : float
        The estimated mutual information between X and Y.
    """
    if np.size(X) == 0 or np.size(Y) == 0:
        return 0
    X = np.atleast_2d(X.T).T
    Y = np.atleast_2d(Y.T).T
    X = check_array(X, ensure_min_samples=2)
    Y = check_array(Y, ensure_min_samples=2)
    assert len(X) == len(Y), 'X and Y must have the same length.'

    mi_estimator = kwargs.get('mi_estimator', 'knn')
    if mi_estimator == 'dr':
        sigma = kwargs.get('sigma', 1)
        n_bases = kwargs.get('n_bases', 200)
        maxiter = kwargs.get('maxiter', 1000)
        mi = mi_dr(X, Y, sigma=sigma, n_bases=n_bases, maxiter=maxiter)
    elif mi_estimator == 'knn':
        k = kwargs.get('k', 3)
        assert isinstance(k, (int, np.integer)) and k > 0, 'k must be a positive integer.'
        assert k < len(X), '`k` must be smaller than `n_sample`.'
        mi = mi_knn(X, Y, k)
    else:
        raise NotImplementedError

    return max(mi, 0) if force_non_negative else mi


def conditional_mutual_information(X, Y, Z, force_non_negative=False, **kwargs):
    """Estimate conditional mutual information for discrete-continuous mixutres.

    Parameters
    ----------
    X : array-like, shape (n_samples, d_x)
        Conditioned variable.
    Y : array-like, shape (n_samples, d_y)
        The other conditioned variable.
    Z : array-like, shape (n_samples, d_z)
        Conditioning variable.
    force_non_negative : bool
        If True, the result will be taken max with zero. Default is False.
    kwargs : dict, default None
        Optional parameters for MI estimation.

    Returns
    -------
    cmi : float
        Estimated conditional mutual information between each X and Y, given Z.
    """
    if np.size(Z) == 0:
        return mutual_information(X, Y, force_non_negative, **kwargs)
    assert len(X) == len(Y) == len(Z), 'X, Y and Z must have the same length.'
    X = np.atleast_2d(X.T).T
    Z = np.atleast_2d(Z.T).T
    XZ = np.hstack([X, Z])
    mi_xz_y = mutual_information(XZ, Y, force_non_negative, **kwargs)
    mi_y_z = mutual_information(Y, Z, force_non_negative, **kwargs)
    cmi = mi_xz_y - mi_y_z
    return max(cmi, 0) if force_non_negative else cmi


def mimat(X, force_non_negative=False, **kwargs):
    """Dimension-wise mutual information.

    Parameters
    ----------
    X : array-like, shape (n_samples, d)
        The observations of a set of variables.
    force_non_negative : bool
        If True, the result will be taken max with zero. Default is False.
    kwargs : dict, default None
        Optional parameters for MI estimation.

    Returns
    -------
    mis : array, shape (d, d)
        Estimated pairwise MIs between dimensions.
    """
    n, d = X.shape
    mis = np.eye(d)
    for i, j in [(i, j) for i in range(d) for j in range(i + 1, d)]:
        x = X[:, [i]]
        y = X[:, [j]]
        mis[i, j] = mis[j, i] = mutual_information(x, y, force_non_negative, **kwargs)
    return mis


def cmimat(X, force_non_negative=False, **kwargs):
    """Dimension-wise conditional mutual information.

    Parameters
    ----------
    X : array-like, shape (n_samples, d)
        The observations of a set of variables.
    force_non_negative : bool
        If True, the result will be taken max with zero. Default is False.
    kwargs : dict, default None
        Optional parameters for MI estimation.

    Returns
    -------
    cmis : array, shape (d, d)
        Estimated pairwise CMIs between dimensions, given the other dimensions.
    """
    n, d = X.shape
    cmis = np.eye(d)
    for i, j in [(i, j) for i in range(d) for j in range(i + 1, d)]:
        x = X[:, [i]]
        y = X[:, [j]]
        idx_rest = (np.arange(d) != i) & (np.arange(d) != j)
        z = X[:, idx_rest]
        cmis[i, j] = cmis[j, i] = conditional_mutual_information(x, y, z, force_non_negative, **kwargs)
    return cmis



import numpy as np
from scipy.stats import multivariate_normal
from scipy.optimize import minimize
from sklearn.cluster import KMeans


def mi_dr(X, Y, sigma, n_bases, maxiter):
    """Estimate mutual information between X and Y using density ratio estimation.

    Parameters
    ----------
    X : array-like, shape (n_samples, d_x) or (n_samples)
        The observations of a variable.
    Y : array-like, shape (n_samples, d_y) or (n_samples)
        The observations of the other variable.
    sigma : float
        The kernel width for density ratio estimator.
    n_bases : int
        The number of bases used in density ratio estimation.
    maxiter : int
        The maximum number of iteration in density ratio estimation.

    Returns
    -------
    mi : float
        The estimated mutual information between X and Y.
    """
    n, d_x = X.shape
    _, d_y = Y.shape
    b = min(n_bases, n)
    XY = np.hstack([X, Y])
    UV = KMeans(b).fit(XY).cluster_centers_
    U, V = np.split(UV, [d_x], axis=1)
    phi_x = np.array([_normal(X, u, sigma) for u in U])
    phi_y = np.array([_normal(Y, v, sigma) for v in V])
    phi = phi_x * phi_y
    h_x = np.sum(phi_x, axis=1)
    h_y = np.sum(phi_y, axis=1)
    h_xy = np.sum(phi, axis=1)
    h = (h_x * h_y - h_xy) / (n ** 2 - n)

    def fun(alpha):
        return -np.sum(np.log(alpha.dot(phi)))

    def jac(alpha):
        return -phi.dot(1 / alpha.dot(phi))

    x0 = np.random.uniform(0, 1, b)
    bounds = [(0, None)] * b
    constraints = [{'type': 'eq', 'fun': lambda alpha: alpha.dot(h) - 1}]
    result = minimize(fun=fun, jac=jac, x0=x0, bounds=bounds, constraints=constraints, options={'maxiter': maxiter})

    if not result.success:
        raise Warning('Optimization failed: %s' % result.message)

    mi = np.mean(np.log(result.x.dot(phi)))
    return mi


def _normal(X, mean, sigma):
    cov = sigma ** 2 * np.eye(len(mean))
    return multivariate_normal.pdf(X, mean=mean, cov=cov)



from depynd.information._dr import mi_dr
from depynd.information._knn import mi_knn
from depynd.information._information import mutual_information, conditional_mutual_information, mimat, cmimat

__all__ = ['mi_dr', 'mi_knn', 'mutual_information', 'conditional_mutual_information', 'mimat', 'cmimat']



from setuptools import setup, find_packages

long_description = """`depynd` is a Python library for learning dependencies between random variables from data.
 It supports
- Mutual information (MI) estimation for discrete-continuous mixtures
- Feature selection algorithms based on MI
- Markov random field structure learning algorithms
"""

setup(
    name='depynd',
    version='0.4.5',
    description='Evaluating dependencies among random variables.',
    long_description=long_description,
    author='Yuya Takashina',
    author_email='takashina2051@gmail.com',
    packages=find_packages(),
    install_requires=[
        'numpy', 'scipy', 'scikit-learn',
    ],
    url='https://github.com/y-takashina/depynd',
)



import numpy as np

from depynd.information import conditional_mutual_information


def mifs(X, y, lamb=0.0, **kwargs):
    """Select effective features in X on predinting y using mutual-information-based feature selection [1]_.

    Parameters
    ----------
    X : array-like, shape (n_samples, d)
        Features.
    y : array-like, shape (n_samples)
        Target. Can be either continuous or discrete.
    lamb: float
        Threshold for independence tests.
    kwargs : dict, default None
        Optional parameters for MI estimation.

    Returns
    -------
    indices : list
        Indices for the selected features.

    References
    ----------
    .. [1] Brown, Gavin, et al. "Conditional likelihood maximisation: a unifying framework for information theoretic
    feature selection." Journal of machine learning research 13.Jan (2012): 27-66.
    """
    n, d = X.shape
    selected = []
    while True:
        max_cmi = -np.inf
        not_selected = set(range(d)) - set(selected)
        z = X[:, selected]
        for i in not_selected:
            x = X[:, i]
            cmi = conditional_mutual_information(x, y, z, **kwargs)
            if max_cmi < cmi:
                max_cmi = cmi
                max_idx = i
        if max_cmi <= lamb or len(selected) == d:
            break
        selected.append(max_idx)

    while True:
        min_cmi = np.inf
        for i in selected:
            x = X[:, i]
            z = X[:, list(set(selected) - {i})]
            cmi = conditional_mutual_information(x, y, z, **kwargs)
            if min_cmi > cmi:
                min_cmi = cmi
                min_idx = i

        if min_cmi >= lamb or len(selected) == 0:
            return selected

        selected.remove(min_idx)



from sklearn.utils.validation import check_X_y

from depynd.feature_selection import mrmr, mifs


def select(X, y, lamb=0.0, method='mifs', **kwargs):
    """Select effective features in X on predinting y using

    Parameters
    ----------
    X : array-like, shape (n_samples, d)
        Features.
    y : array-like, shape (n_samples)
        Target. Can be either continuous or discrete.
    lamb: float, default 0.0
        Threshold for independence tests.
    method: str, default 'mifs'
        Method used for feature selection. Either 'mifs' or 'mrmr' can be chosen.

    Returns
    -------
    indices : list
        Indices for the selected features.
    """
    X, y = check_X_y(X, y)
    if method == 'mifs':
        return mifs(X, y, lamb=lamb, **kwargs)
    elif method == 'mrmr':
        return mrmr(X, y, lamb=lamb, **kwargs)
    else:
        raise NotImplementedError('Method %s is not implemented. Use mifs or mrmr.' % method)
